/*
  parseContig.c

  $Id: parseContig.c,v 1.1.1.1 2005/08/04 14:32:29 mcschatz Exp $

  parse multi-contig .contig files
  
*/

#include "getCoverage.h"

FILE * f_contig = NULL;

void readContig(FILE *f, struct CONTIG_SEQUENCE **pcontig, int *numOfSeqs)
{
  char c;
  int gAsmPos = 0;
  int contigLen;
  char name[MAXSEQNAME];
  char checksum[9];
  struct CONTIG_SEQUENCE *contig; 
  int numOfGaps = 0;

  f_contig = f;

  if (fscanf(f,"##%s %d %d bases, %s checksum.",
         name, numOfSeqs, &contigLen, checksum) != 4)
  {
    printError("readContig", DEBUG_ERROR,
               "ERROR: Invalid contig header line!");

    exit(1); // invalid contig header line
  }

  contig = (struct CONTIG_SEQUENCE *) newmem2(1,sizeof(struct CONTIG_SEQUENCE), 
                                               __FILE__,__LINE__);
  *pcontig = contig;

  contig->len = contigLen;

  contig->id = (char *) newmem2(strlen(name)+1,sizeof(char), __FILE__,__LINE__);
  strcpy(contig->id,name);

  contig->seq = (char *) newmem2(contigLen+1, sizeof(char), __FILE__,__LINE__);

  // Read sequence into allocated space.
  while ((c=fgetc(f)) != EOF && c != '#')
  {
    if (!isspace(c)) 
    { 
      if (gAsmPos >= contigLen)
      {
        printError("readContig", DEBUG_ERROR,
                   "ERROR: Consensus of %s is longer than reported (%d)!",
                   contig->id, contigLen);

        exit(1); // consensus is too wide
      }

      if (c == '-') { numOfGaps++; }

      contig->seq[gAsmPos] = c; 
      gAsmPos++;
    } 
  }

  contig->seq[gAsmPos] = '\0'; // make it easy to print entire sequence

  // remember numOfGaps so we know the ungapped length of the consensus
  // but don't bother find the gapInd
  contig->numOfGaps = numOfGaps; 
  contig->gapInd = NULL;
  
  // Make sure f_contig is in position to read the first sequence record
  ungetc(c,f); // c == EOF || c == '#'
}

int seqoffsetcmp(const void * a, const void * b)
{
  const struct CONTIG_SEQUENCE * ac = *(const struct CONTIG_SEQUENCE **)a;
  const struct CONTIG_SEQUENCE * bc = *(const struct CONTIG_SEQUENCE **)b;

  int retval = ac->offset - bc->offset;
  
  if (retval == 0)
  {
    retval = ac->len - bc->len;

    if (retval == 0)
    {
      retval = strcmp(ac->id,bc->id);
    }
  }

  return retval;
}

struct CONTIG_SEQUENCE_LIST
{
  struct CONTIG_SEQUENCE * cseq;
  struct CONTIG_SEQUENCE_LIST * next;
};

/*
  readSeqs()

  parse reads information from .contig file
  Only extract sequences that overlap the range of interest
*/
char readSeqs(int * numOfSeqs,
              struct CONTIG_SEQUENCE ***bcseq,
              struct CONTIG_SEQUENCE * contig,
              Coordinates * coords,
              const char * seqname_range,
              char assumeCircular,
              char loadAll)   
{
  off_t f_offset;

  int namelen;
  int len;
  int reportedlen;
  int offset;
  int numGaps;
  int seql, seqr, asmr, asml;

  int readsSeen = 0;
  int readcount = 0;
  int readid;

  char name[MAXSEQNAME];          // name string buffer input 
  char checksum[9];               // space for checksum word in contig file
  char c, c2=0;
  char foundseqname = 0;
  char rc;

  struct CONTIG_SEQUENCE ** cseqarr = NULL; // array of all
  struct CONTIG_SEQUENCE_LIST * cseqlist = NULL; // linked list head
  struct CONTIG_SEQUENCE_LIST * templist = NULL; // linked list node

  c = fgetc(f_contig);
  while (c != EOF && !(c == '#' && (c2=fgetc(f_contig)) == '#'))
  {
    readsSeen++;
    numGaps = 0;

    // Read the sequence name into the name array 
    c = c2; // c was '#'
    for (namelen=0; c != '('; namelen++, c = fgetc(f_contig)) 
    {
      if (c == EOF)
      {
        printError("parseContig", DEBUG_ERROR, "contig is invalid");
        exit(1); // invalid contig;
      }

      name[namelen] = c;
    }
    name[namelen] = '\0';  

    // read the offset 
    fscanf(f_contig,"%d) [", &offset);
    
    // read reverse complement flag 
    c = fgetc(f_contig);
    if (c == 'R') { rc = 1;  fgetc(f_contig);    }
    else          { rc = 0; ungetc(c, f_contig); } // c == ']'

    fscanf(f_contig, "] %d bases, %s checksum. {%d %d} <%d %d>",
           &reportedlen, checksum, &seql, &seqr, &asml, &asmr);

    f_offset = ftello(f_contig);

    // Count the actual length of the sequence and number of gaps
    len = 0;
    while (c != EOF && c != '#')
    {
      c = fgetc(f_contig);
      if (c != EOF && c != '#' && !isspace(c)) 
      { 
        len++; 
        if (c == '-') { numGaps++; }
      }
    }

    if (len != reportedlen)
    {
      printWarning("readSeqs",
                   "Read %s is %d bases, but is reported to be %d bases",
                   name, len, reportedlen);

    }

    if (offset + len > contig->len)
    {
      int delta = offset+len - contig->len;
      if (assumeCircular)
      {
        // Normalize and give a negative offset
        offset = offset - contig->len;
        printWarning("parseContig",
                     "Read %s extends beyond end of contig by %d bp, assuming circular read",
                     name, delta);
      }
      else
      {
        // After some consideration, by default warn about read, then truncate
        len = contig->len - offset;
        printWarning("parseContig",
                     "Read %s extends beyond end of contig by %d bp, truncating",
                     name, delta);
        printWarning("parseContig",
                     "  Use -C if the read should be intrepreted as circular");
      }
    }
    
    // Check to see if this sequences is within range of interest
    // Grab everybody if seqname_range so we can prune after determine real range
    if (loadAll || hasOverlap(coords->start, coords->end-1, 
                              offset, len, contig->len))
    {
      struct CONTIG_SEQUENCE * cseq = (struct CONTIG_SEQUENCE *) newmem2(1, sizeof(struct CONTIG_SEQUENCE),
                                                __FILE__,__LINE__);

      templist = pool_malloc(sizeof(struct CONTIG_SEQUENCE_LIST));
      templist->next = cseqlist;
      templist->cseq = cseq;
      cseqlist = templist;

      readcount++;

      cseq->rc = rc;
      cseq->len = len;
      cseq->offset = offset;
      cseq->numOfGaps = numGaps;

      cseq->seq_l = seql;
      cseq->seq_r = seqr;
      cseq->asm_l = asml;
      cseq->asm_r = asmr;

      cseq->f_offset = f_offset;

      cseq->id = (char *)newmem2(namelen+1,sizeof(char),__FILE__,__LINE__);
      strcpy(cseq->id,name);

      if (seqname_range && !foundseqname && !strcmp(seqname_range, cseq->id))
      {
        foundseqname = 1;
        setRangeCoordinates(cseq, contig, coords);
      }

      /*
      fprintf(stderr, "%s [%d %d] overlaps [%d %d]\n",
              cseq->id, offset, offset+len-1, start, end);
      */
    }
  }

  if(c != EOF && c2 != EOF)
  {
    ungetc(c2, f_contig); // c2 == '#' before contig record
    ungetc(c, f_contig);  // c1 == '#' before contig record
  }

  if (readsSeen != *numOfSeqs)
  {
    printWarning("readSeqs",
                 "contig %s has %d sequences but reported to have %d",
                 contig->id, readsSeen, *numOfSeqs);
  }
  
  if (seqname_range)
  {
    if (!foundseqname)
    {
      // Sequence not found
      printError("readSeqs", DEBUG_ERROR,
                 "ERROR: sequence %s not found", seqname_range);
      exit(1); // seqname_range not found
    }

    if (!loadAll)
    {
      readcount = 0;

      // drop elements from the head
      while (cseqlist && !hasOverlap(coords->start, coords->end-1, 
                                     cseqlist->cseq->offset, 
                                     cseqlist->cseq->len,
                                     contig->len))
      {
        free(cseqlist->cseq->id);
        free(cseqlist->cseq);
        cseqlist = cseqlist->next;
      }

      if (cseqlist)
      {
        struct CONTIG_SEQUENCE_LIST * nextnext;

        readcount = 1; // cseqlist->cseq is an overlapping read
        templist = cseqlist;
      
        // templist now points at a valid node (or NULL)
        while (templist->next)
        {
          if (!hasOverlap(coords->start, coords->end-1,
                          templist->next->cseq->offset, 
                          templist->next->cseq->len,
                          contig->len))
          {
            nextnext = templist->next->next;
            free(templist->next->cseq->id);
            free(templist->next->cseq);

            templist->next = nextnext;
            // don't advance templist since nextnext may need to be dropped too
          }
          else
          {
            templist = templist->next;
            readcount++; //templist->next is an overlapping read
          }
        }
      }
    }
  }

  // Flatten the linked list into an array (now that we know the real count)
  cseqarr = (struct CONTIG_SEQUENCE **) newmem2(readcount,
                                                sizeof(struct CONTIG_SEQUENCE *),
                                                __FILE__,__LINE__);

  for(readid = 0; readid < readcount; readid++)
  {
    cseqarr[readid] = cseqlist->cseq;
    cseqlist = cseqlist->next; 
  }

  pool_flush(); // cleanup the linked list

  // Now sort the sequences by offset
  qsort(cseqarr, readcount, sizeof(struct CONTIG_SEQUENCE *), seqoffsetcmp);
 
  *bcseq = cseqarr;
  *numOfSeqs = readcount;
  
  return c;
}

void setRangeCoordinates(struct CONTIG_SEQUENCE * bcseq,
                         struct CONTIG_SEQUENCE * contig,
                         Coordinates * coords)
{
  // Use this read's coordinates as the start and end coordinates
  coords->start = bcseq->offset;
  coords->end = bcseq->offset + bcseq->len; // 1 past end
  coords->ugInd = bcseq->asm_l-1;          // switch to 0-based

  if (coords->start < 0)
  {
    printWarning("setRangeCoordinates",
                 "Read %s is circular, using entire contig",
                 bcseq->id);

    coords->ugInd = 0;
    coords->start = 0;
    coords->end = contig->len;
  }

  logMsg(DEBUG_LOW, "getCoverage",
         "Reseting range to [%d, %d] %d\n",
         coords->start, coords->end-1, coords->ugInd);
}



/*
  readSeqs()

  parse reads information from .contig file
  Only extract sequences that overlap the range of interest
*/
void loadSeq(struct CONTIG_SEQUENCE *bcseq)
{
  char c;
  int i = 0;
  int currentGap = 0;
  
  // Allocate memory to hold sequence 
  // The length has already been verified, except reads that hang off the end
  bcseq->seq = (char *) pool_malloc((bcseq->len)+1);

  // Allocate space for the gaps which have already been counted
  bcseq->gapInd = (int *) pool_malloc(bcseq->numOfGaps * sizeof(int));

  fseeko(f_contig, bcseq->f_offset, SEEK_SET);

  // Read sequence into allocated space. 
  while ((c=fgetc(f_contig)) != EOF && c != '#' && i < bcseq->len)
  {
    if (!isspace(c)) 
    { 
      bcseq->seq[i] = c; 

      if (c == '-')
      {
        bcseq->gapInd[currentGap] = i-currentGap;
        currentGap++;
      }

      i++;
    } 
  }

  bcseq->seq[i] = '\0';
  bcseq->numOfGaps = currentGap;

  // now, gapInd[k] is a position in sequence (ungapped based at 0)
  // of an element after which a gap is found in lseq
  // for example if lseq = ac-c--t, then seq = acct and gaps = (1,2,2)
}





int hasOverlap(int rangeStart, // 0-based exact offset of range
               int rangeEnd,   // 0-based exact end of range
               int seqOffset,  // 0-bases exact offset of seq
               int seqLen,     // count of bases of seq (seqend+1)
               int contigLen)  // count of bases in contig (contigend+1)
{
  int retval = 1;

  if (seqOffset >= 0)
  {
    if ((seqOffset > rangeEnd)                  // sequence right flanks
        || (seqOffset + seqLen-1 < rangeStart)) // sequence left flanks
    {
      retval = 0;
    }
  }
  else
  {
    // Test left and right separately
    retval = hasOverlap(rangeStart, rangeEnd, 
                        0, seqLen+seqOffset, contigLen) ||
             hasOverlap(rangeStart, rangeEnd, 
                        contigLen+seqOffset, -seqOffset, contigLen);
  }

  return retval;
}
