#include "ConsensusSequence.hh"

ConsensusSequence::ConsensusSequence(int contigCount, 
                                     long index, 
                                     ostream & out)
  : m_out(out)
{
  m_contigCount = contigCount;
  m_asmleft = index;

  m_curOffset = 0;
  m_coordinateOffset = 1-index;
}

ConsensusSequence::~ConsensusSequence()
{
  SequenceList::const_iterator i;
  for (i =  m_completeSequenceList.begin();
       i != m_completeSequenceList.end();
       i++)
  {
    delete *i;
  }
}

// Flush sequences that were not seen in the last pass
void ConsensusSequence::flushSequences()
{
  SequenceMap::iterator i;
 
  for (i =  m_previousSequences.begin();
       i != m_previousSequences.end();
       i++)
  {
    i->second->printSequenceXML(m_out, "      ");
  }

  // Set the active sequences to be the previous sequences for the next pass
  m_previousSequences = m_activeSequences;

  // Erase all elements form the activeSequence list
  m_activeSequences.clear();
}

void ConsensusSequence::addSlice(SliceData * slice)
{
  long index = slice->getIndex();

  char consensus = slice->getConsensus();

  addElement(consensus, index);

  unsigned short depth = slice->getDepth();
  for (unsigned short i = 0; i < depth; i++)
  {
    SequenceData * sd = getSequence(slice->getRead(i), m_curOffset, index);

    sd->addElement(index, slice->getBase(i), slice->getQualVal(i));

    addActiveSequence(sd);
  }

  m_curOffset++;

  flushSequences();
}

void ConsensusSequence::addElement(char base, long index)
{
  m_nuc += base;
  m_asmright = index;
}


SequenceData * ConsensusSequence::getSequence(const ReadData * rd, 
                                              long offset,
                                              long index)
{
  SequenceData * retval;
  SequenceMap::const_iterator i;

  i = m_completeSequenceMap.find(rd->getId());

  if (i == m_completeSequenceMap.end())
  {
    retval = addSequence(rd, offset, index);
  }
  else
  {
    retval = i->second;
  }

  return retval;
}

SequenceData * ConsensusSequence::addSequence(const ReadData * rd,
                                              long offset,
                                              long index)
{
  SequenceData * retval = new SequenceData(rd, offset, index);
  m_completeSequenceMap[rd->getId()] = retval;
  m_completeSequenceList.push_back(retval);
  
  return retval;
}

void ConsensusSequence::addActiveSequence(SequenceData * sd)
{
  // First add the sequence to the map of active sequences
  const string & seqid = sd->getId();
  m_activeSequences[seqid] = sd;

  // Now Remove this sequence from the map of previous sequences
  SequenceMap::iterator i = m_previousSequences.find(seqid);
  if (i != m_previousSequences.end())
  {
    m_previousSequences.erase(i);
  }
}

void ConsensusSequence::printTileXML(const string & prefix)
{
  m_out << prefix << "<Contig Id=\"ASM_" << m_contigCount << "\">" << endl;
  m_out << prefix << "  <Nuc>" << m_nuc << "</Nuc>" << endl;

  SequenceList::const_iterator i;
  for (i =  m_completeSequenceList.begin();
       i != m_completeSequenceList.end();
       i++)
  {
    (*i)->printSeqRecordXML(m_out, prefix + "  ", m_coordinateOffset);
  }

  m_out << prefix << "</Contig>" << endl;
}
